package moreswing.swing

import scala.collection.Iterator
import scala.collection.mutable.Buffer

/** Default partial implementation for buffer adapters.
 *
 * @author myst3r10n
 */
protected[moreswing] abstract class BufferWrapper[A] extends Buffer[A] { outer =>

  def clear() { for (i <- 0 until length) remove(0) }

  def update(n: Int, a: A) {
    remove(n)
    insertAt(n, a)
  }

  def insertAll(n: Int, elems: Traversable[A]) {
    var i = n
    for (el <- elems) {
      insertAt(i, el)
      i += 1
    }
  }

  protected def insertAt(n: Int, a: A)
  
  def +=:(a: A): this.type = { insertAt(0, a); this }
  def iterator = Iterator.range(0,length).map(apply(_))

}

