package moreswing.swing

import event. {

  InternalFrameActivated,
  InternalFrameClosed,
  InternalFrameMaximized,
  InternalFrameUnmaximized

}

import event.PageDetached

import java.awt.event.MouseEvent
import java.beans. { PropertyChangeEvent, PropertyChangeListener }

import javax.swing. { Icon, ImageIcon, JTabbedPane, WindowConstants }
import javax.swing.event. { ChangeEvent, ChangeListener }

import scala.collection.mutable. { Buffer, Map }

import scala.swing. {

  Alignment,
  Button,
  Color,
  Component,
  FlowPanel,
  Frame,
  Label,
  Rectangle,
  Publisher

}

import scala.swing.event. {

  ButtonClicked,
  MouseDragged,
  MouseMoved,
  MousePressed,
  MouseReleased,
  SelectionChanged,
  WindowClosing

}

import moreswing.swing.plaf.Redesign

/** Companion of [[scala.swing.TabbedDesktopPane]].
 *
 * @author myst3r10n
 */
object TabbedDesktopPane {

  class Page protected[TabbedDesktopPane](parent0: TabbedDesktopPane,
                                          title0: String, icon0: Icon,
                                          content0: Component, tip0: String) extends Proxy {

    def self = content0

    def this(title0: String, icon0: ImageIcon,
             content0: Component, tip0: String) =
             this(null, title0, icon0, content0, tip0)

    def this(title0: String, icon0: ImageIcon,
             content0: Component) = this(title0, icon0, content0, "")

    protected var _border = new BorderPanel

    content = content0
    title = title0
    icon = icon0
    tip = tip0

    protected[TabbedDesktopPane] var parent: TabbedDesktopPane = parent0

    protected var _background: Color = null

    /** @see [[javax.swing.JTabbedPane]]. */
    def background = _background

    /** @see [[javax.swing.JTabbedPane]]. */
    def background_=(c: Color) {

      _background = c

      if(parent != null)
        parent.peer.setBackgroundAt(index, c)

    }

    /** @see [[javax.swing.JTabbedPane]]. */
    def bounds: Rectangle =
      if(parent == null)
        null
      else
        parent.peer.getBoundsAt(index)

    def border = _border

    protected def border_=(bp: BorderPanel) {

      _border = bp

      if(parent != null)
        parent.peer.setComponentAt(index, bp.peer)

    }

    protected var _content = content0

    /** @see [[javax.swing.JTabbedPane]]. */
    def content = _content

    /** @see [[javax.swing.JTabbedPane]]. */
    def content_=(c: Component) {

      if(c != null) {

      _content = c

      if(parent != null && parent.iframes.contains(index))
        parent.iframes(index).contents = c
      else
        border.add(c, scala.swing.BorderPanel.Position.Center)

      }
    }

    protected var _enabled = true

    /** @see [[javax.swing.JTabbedPane]]. */
    def enabled = _enabled

    /** @see [[javax.swing.JTabbedPane]]. */
    def enabled_=(b: Boolean) {

      _enabled = b

      if(parent != null)
        parent.peer.setEnabledAt(index, b)

    }

    protected var _foreground: Color = null

    /** @see [[javax.swing.JTabbedPane]]. */
    def foreground = _foreground

    /** @see [[javax.swing.JTabbedPane]]. */
    def foreground_=(c: Color) {

      _foreground = c

      if(parent != null) {
        parent.peer.setForegroundAt(index, c)
        val tab = parent.peer.getTabComponentAt(index)
        if(tab != null)
          tab.asInstanceOf[javax.swing.JPanel].getComponents()(0)
            .asInstanceOf[javax.swing.JLabel].setForeground(c)

      }

    }

    protected var _icon = icon0

    /** @see [[javax.swing.JTabbedPane]]. */
    def icon = _icon

    /** @see [[javax.swing.JTabbedPane]]. */
    def icon_=(i: Icon) {

      _icon = i

      if(parent != null) {

        parent.peer.setIconAt(index, i)
        parent.peer.getTabComponentAt(index)
          .asInstanceOf[javax.swing.JPanel].getComponents()(0)
          .asInstanceOf[javax.swing.JLabel].setIcon(i)

        if(parent.iframes.contains(index))
          parent.iframes(index).frameIcon = i

      }
    }

    /** @see [[javax.swing.JTabbedPane]]. */
    def index =
      if(parent == null)
        -1
      else
        parent.peer.indexOfComponent(border.peer)

    protected var _mnemonic = -1

    /** @see [[javax.swing.JTabbedPane]]. */
    def mnemonic = _mnemonic

    /** @see [[javax.swing.JTabbedPane]]. */
    def mnemonic_=(k: Int) {

      _mnemonic = k

      if(parent != null)
        parent.peer.setMnemonicAt(index, k)

    }

    protected var _tip = tip0

    /** @see [[javax.swing.JTabbedPane]]. */
    def tip = _tip

    /** @see [[javax.swing.JTabbedPane]]. */
    def tip_=(t: String) {

      _tip = t

      if(parent != null)
        parent.peer.setToolTipTextAt(index, t)

    }

    protected var _title = title0

    /** @see [[javax.swing.JTabbedPane]]. */
    def title = _title

    /** @see [[javax.swing.JTabbedPane]]. */
    def title_=(t: String) {

      _title = title

      if(parent != null) {

        parent.peer.setTitleAt(index, t)
        parent.peer.getTabComponentAt(index)
          .asInstanceOf[javax.swing.JPanel].getComponents()(0)
          .asInstanceOf[javax.swing.JLabel].setText(t)

        if(parent.iframes.contains(index))
          parent.iframes(index).title = t

      }
    }
  }
}

/** Like [[scala.swing.TabbedPane]] but with draggable, detachable, reattachable, arrangeable and closeable tabs.
 *
 * @author myst3r10n
 */
class TabbedDesktopPane extends Component with Publisher {

  override lazy val peer: JTabbedPane = new JTabbedPane with SuperMixin

  import TabbedDesktopPane.Page
  import scala.swing.TabbedPane.Layout

  protected var _closeable = true

  def closeable = _closeable

  def closeable_= (b: Boolean) {

    _closeable = b
    rebuildTabBars

  }

  protected var _detachable = true

  def detachable = _detachable

  def detachable_= (b: Boolean) {

    _detachable = b
    rebuildTabBars

  }

  protected var _shiftable = true

  /** Shift mode. */
  def shiftable = _shiftable

  /** Toggle shift mode.
   *
   * @param b If this shift mode is enabled.
   */
  def shiftable_= (b: Boolean) { _shiftable = b }

  /** Shift mode flotable. */
  def flotableShift = desktop.flotableShift

  /** Toggle shift mode flotable.
   *
   * @param b If this shift mode is flotable.
   */
  def flotableShift_= (b: Boolean) { desktop.flotableShift = b }

  /** Desktop mode gridable. */
  def gridableDesktop = desktop.gridableDesktop

  /** Toggle desktop mode gridable.
   *
   * @param b If this desktop mode is gridable.
   */
  def gridableDesktop_= (b: Boolean) {

    // Disable maximum mode.
    maximum = false

    desktop.gridableDesktop = b

  }

  /** @see [[scala.swing.TabbedPane]]. */
  object pages extends Buffer[Page] {

    /** Appends a tab to this buffer.
     *
     * @param t The tab to append.
     *
     * @return The buffer itself.
     */
    def +=(t: Page): this.type = {

      insert(if(size < 1) 0 else size, t)
      this

    }

    /** Prepends a tab to this buffer.
     *
     * @param t The tab to prepend.
     *
     * @return The buffer itself.
     */
    def +=:(t: Page): this.type = {

      insert(0, t)
      this

    }

    /** Select a tab by its index in the buffer.
     *
     * @param n The index where tabs are selected.
     *
     * @return The tab at index n.
     */
    def apply(n: Int): Page = {

      new Page(TabbedDesktopPane.this, peer.getTitleAt(n),
               peer.getIconAt(n), null, peer.getToolTipTextAt(n)) {

        border = new BorderPanel {
          override lazy val peer = TabbedDesktopPane.this.peer
            .getComponentAt(n).asInstanceOf[javax.swing.JPanel with SuperMixin]
        }

        content = 
          if(iframes.contains(n))
            iframes(n).contents.headOption.orNull
          else
            border.layoutManager.getLayoutComponent(scala.swing.BorderPanel.Position.Center.toString).asInstanceOf[javax.swing.JComponent] match {

                case null => null
                case peer: javax.swing.JComponent =>
                  peer.getClientProperty("scala.swingWrapper").asInstanceOf[Component]

            }
      }
    }

    /** Clears the buffer's contents. */
    def clear {

      for(i <- 0 to size - 1)
        remove(i)

    }

    /** Detach tab at given index.
     * 
     * @param n The index where tabs are detached.
     *
     * @return The detached frame.
     */
    def detach(n: Int) = {
    
      val frame = new Frame with Redesign {

        // Save tab.
        val tab = pages(n)

        // Use better close operation.
        peer.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE)

        // Set look.
        title = tab.title
        if(pages(n).icon != null)
          iconImage = pages(n).icon.asInstanceOf[ImageIcon].getImage

        // Move tab content to frame.
        if(maximum)
          contents = tab.border
        else {

          iframes(tab.index).contents.headOption.orNull match {
            case content: Component =>
              tab.border.add(content, scala.swing.BorderPanel.Position.Center)
          }

          relations.unbind(tab.index).closed = true
          contents = tab.border

          if(pages.length > 0)
            selection.page.border.add(desktop,
                                      scala.swing.BorderPanel.Position.Center)

        }

        // Event handling.
        var reattached = false
        reactions += {
          case WindowClosing(_) => 

            if(!reattached) {

              reattached = true
             	  reattach(tab)

          }
        }

        pack
        visible = true

      }

      TabbedDesktopPane.this.publish(PageDetached(TabbedDesktopPane.this, frame))

      frame

    }

    /** Arranges the internal frame horizontal. */
    def horizontal { 

      // Disable maximum mode.
      TabbedDesktopPane.this.maximum = false

      // Arrange horizontal.
      desktop.horizontal

      // Select current tab.
      iframes(selection.index).selected = true

    }

    /** Insert tab at given index into this buffer.
     *
     * @param n The index where new tab are inserted.
     * @param t The tab to insert.
     */
    def insert(n: Int, t: Page) {

      if(!maximum)
        maximum = true

      // Set parent.
      t.parent = TabbedDesktopPane.this

      // Inserts tab and override tab bar.
      peer.insertTab(t.title, t.icon, t.border.peer, t.tip, n)
      if(t.background == null)
        t.background = peer.getBackgroundAt(t.index)
      else
        peer.setBackgroundAt(t.index, t.background)
      peer.setEnabledAt(t.index, t.enabled)
      overrideTabBar(t)
      t.foreground = peer.getForegroundAt(t.index)

      // Set current tab.
      selection.index = n

    }

    /** Inserts new tabs at a given index into this buffer.
     *
     * @param n The index where new tabs are inserted.
     * @param tabs The tabs to insert.
     */
    def insertAll(n: Int, tabs: Traversable[Page]) {

      var i = n
      for(t <- tabs) {

        insert(i, t)
        i += 1

      }
    }

    def iterator = Iterator.range(0, length).map(apply(_))

    /** The length of the buffer.
     *
     * @return The length of the buffer.
     */
    def length = peer.getTabCount

    /** Reattach the frame.
     *
     * @param t The detached tab.
     */
    def reattach(t: Page) { pages += t }

    /** Removes the tab at a given index from this buffer.
     *
     * @param n The index which refers to the tab to delete.
     *
     * @return The previous tab at index n.
     */
    def remove(n: Int): Page = {

      val t = apply(n)

      // Release parent.
      t.parent = null

      // Removes tab.
      if(maximum)
        peer.removeTabAt(n)
      else
        iframes(n).closed = true

      t

    }

    /** Removes inatvice tabs from this buffer. */
    def removeInactive {

      shift(selection.index, 0)

      var i = 1
      while(i < size)
        remove(i)

    }

    /** Shifts tab within buffer.
     *
     * @param from The index of tab.
     * @param to The index where tab should be shifted.
     */
    def shift(from: Int, to: Int) {

      val t = apply(from)

      // Removes tab.
      selection.skip = true
      peer.removeTabAt(from)

      if(!maximum)
        relations.bind(to, relations.unbind(from))

      // Inserts tab and override tab bar.
      if(from > to)
        selection.skip = true
      peer.insertTab(t.title, t.icon, t.border.peer, t.tip, to)
      overrideTabBar(t)

    }

    /** Arranges the internal frame tiling. */
    def tile {

      // Disable maximum mode.
      TabbedDesktopPane.this.maximum = false

      // Arrange tile.
      desktop.tile

      // Select current tab.
      iframes(selection.index).selected = true

    }

    /** Replaces tab at given index with a new tab.
     *
     * @param n The index where tab are updated.
     * @param t The new tab where replaces old tab.
     */
    def update(n: Int, t: Page) {

      remove(n)
      insert(n, t)

    }

    /** Arranges the internal frame vertical. */
    def vertical {

      // Disable maximum mode.
      TabbedDesktopPane.this.maximum = false

      // Arrange tile.
      desktop.vertical

      // Select current tab.
      iframes(selection.index).selected = true

    }
  }

  protected var _popupMenu: PopupMenu = null

  /** Popup menu. */
  def popupMenu = _popupMenu

  /** Set popup menu.
  *
  * @param popupMenu Popup menu.
  */
  def popupMenu_= (popup: PopupMenu) { _popupMenu = popup }

  object selection extends Publisher {

    def page = pages(index)
    def page_=(p: Page) { index = p.index }

    def index = peer.getSelectedIndex
    def index_=(n: Int) { peer.setSelectedIndex(n) }

    protected[TabbedDesktopPane] var skip = false
    protected[TabbedDesktopPane] var previousIndex = -1
    protected[TabbedDesktopPane] var zOrder = Seq[Int]()
    peer.addChangeListener(new ChangeListener {

      def stateChanged(e: ChangeEvent) {

        if(skip)
          skip = false
        else if(!maximum && previousIndex != index) {

          page.border.add(desktop, scala.swing.BorderPanel.Position.Center)

          if(iframes.contains(index) && !iframes(index).selected)
            iframes(index).selected = true

        }

        previousIndex = index

        // Publish that changes happen :)
        publish(SelectionChanged(TabbedDesktopPane.this))

      }
    } )
  }

  /** @see [[javax.swing.JTabbedPane]]. */
  def tabLayoutPolicy = Layout(peer.getTabLayoutPolicy)

  /** @see [[javax.swing.JTabbedPane]]. */
  def tabLayoutPolicy_=(p: Layout.Value) { peer.setTabLayoutPolicy(p.id) }

  /** @see [[javax.swing.JTabbedPane]]. */
  def tabPlacement = Alignment(peer.getTabPlacement)

  /** @see [[javax.swing.JTabbedPane]]. */
  def tabPlacement_=(b: Alignment.Value) { peer.setTabPlacement(b.id) }

  // Event handling.
  private var mouseOverTab = -1
  private var tabDepartFrom = -1
  listenTo(mouse.clicks, mouse.moves, selection)
  reactions += {

    case MouseMoved(_, point, _) =>
      // Save tab where mouse hovers.
      mouseOverTab = peer.indexAtLocation(point.x, point.y)

      // Enable popup menu if mouse over tab.
      if(popupMenu != null &&
         mouseOverTab > -1 &&
         peer.getComponentPopupMenu == null)
        peer.setComponentPopupMenu(popupMenu.peer)

      // Disable popup menu if mouse not over tab.
      else if(mouseOverTab < 0 && peer.getComponentPopupMenu != null)
        peer.setComponentPopupMenu(null)

    case e: MousePressed =>
      // Save tab where mouse hovers.
      mouseOverTab = peer.indexAtLocation(e.point.x, e.point.y)

      // Enable motion mode.
      if(shiftable &&
         e.peer.getButton == MouseEvent.BUTTON1 &&
         tabDepartFrom < 0 &&
         mouseOverTab > -1)
        tabDepartFrom = mouseOverTab

    case e: MouseDragged =>
      // Save tab where mouse hovers.
      mouseOverTab = peer.indexAtLocation(e.point.x, e.point.y)

      val tabArriveIn = mouseOverTab

      // Move tab if flotable shift enabled, destination valid and unequal start.
      if(shiftable &&
         flotableShift &&
         tabDepartFrom > -1 &&
         tabArriveIn > -1 &&
        tabDepartFrom != tabArriveIn) {

          // Shift tab.
          pages.shift(tabDepartFrom, tabArriveIn)
          selection.index = tabArriveIn

          // Sync moved tab.
          tabDepartFrom = tabArriveIn

      }

    case e: MouseReleased =>
      // Save tab where mouse hovers.
      mouseOverTab = peer.indexAtLocation(e.point.x, e.point.y)

      // Pass if button 1 released.
      if(shiftable &&
         e.peer.getButton == MouseEvent.BUTTON1) {

        val tabArriveIn = mouseOverTab

        // Move tab if flotable shift enabled, destination valid and unequal start.
        if(!flotableShift &&
           tabDepartFrom > -1 &&
           tabArriveIn > -1 &&
           tabDepartFrom != tabArriveIn) {

          // Shift tab.
          pages.shift(tabDepartFrom, tabArriveIn)
          selection.index = tabArriveIn
          
          // Sync moved tab.
          tabDepartFrom = tabArriveIn

        }
      }

      // Disable motion mode.
      if(tabDepartFrom != -1)
        tabDepartFrom = -1

  }


  /** Desktop. */
  protected val desktop = new GridDesktopPane

  protected var iframes = Map[Int, InternalFrame]()

  protected object relations {

    /** Bind relation between page and internal frame.
     *
     * @param n The index of tab where refer to internal frame.
     * @param iframe The internal frame.
     */
    def bind(n: Int, iframe: InternalFrame) {

      if(iframes.contains(n)) {

        var i = iframes.size
        while(n < i) {

          iframes += i -> iframes(i - 1)
          i -= 1

        }
      }

      iframes += n -> iframe

    }

    /** Unbind relation between tab and internal frame.
     *
     * @param n The index of tab where refer to internal frame.
     */
    def unbind(n: Int): InternalFrame = {

      val iframe = iframes.remove(n).get
      var i = n
      var last = iframes.size

      while(i < last) {

        iframes += i -> iframes(i + 1)
        i += 1

      }

      if(n < iframes.size)
        iframes -= iframes.size - 1

      iframe

    }

  }

  protected def maximum = iframes.size == 0

  protected def maximum_=(b: Boolean) {

    if(b && iframes.size > 0) {

      for((index, iframe) <- iframes) {

        iframe.contents.headOption.orNull match {
          case content: Component =>
            pages(index).border.add(content, scala.swing.BorderPanel.Position.Center)
        }

        iframes -= index
        iframe.closed = true

      }

      selection.page.border.revalidate

    } else if(!b && iframes.size == 0) {

      val content = selection.page.content
      selection.page.border.add(desktop, scala.swing.BorderPanel.Position.Center)
      selection.page.border.revalidate
      selection.previousIndex = selection.index

      for(tab <- pages.reverse) {

        var skip = true

        // Open relation between tab and internal frame.
        iframes += tab.index -> new InternalFrame(tab.title, true, true, true) {

          // Set default close operation.
          peer.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE)

          // Set icon.
          frameIcon = tab.icon

          // Set content.
          contents = if(tab.index == selection.index) content else tab.content

          // Event handling.
          listenTo(this)
          reactions += {

            case InternalFrameActivated(iframe: InternalFrame) =>
              if(!skip && tab.index != selection.index)
                selection.index = tab.index

            case InternalFrameClosed(iframe: InternalFrame) =>
              if(iframes.find { case (_, iframe) => iframe == this } != None) {

                relations.unbind(tab.index)

                if(tab.index <= selection.index)
                  selection.skip = true

                TabbedDesktopPane.this.peer.removeTabAt(tab.index)

              }

            case InternalFrameMaximized(iframe: InternalFrame) =>
              TabbedDesktopPane.this.maximum = true

          }
        }

        // Adds internal frame to desktop.
        desktop += iframes(tab.index)

        // Show internal frame.
        iframes(tab.index).show

        skip = false

      }
    }
  }

  protected def overrideTabBar(t: Page) {

    // Modify tab component.
    peer.setTabComponentAt(t.index, (new FlowPanel {

      opaque = false

      // Label.
      contents += new Label(t.title, t.icon, Alignment.Left) {

        if(t.foreground == null)
          TabbedDesktopPane.this.peer.setForegroundAt(t.index, foreground)
        else
          foreground = t.foreground

      }

      // Detach Button.
      if(detachable)
        contents += new Button {

          border = null
          icon = new ImageIcon(getClass.getResource("/icons/tab_detach.png"))
          opaque = false

          listenTo(this)
          reactions += {
            case ButtonClicked(_) =>
              pages.detach(t.index)
          }

          peer.addPropertyChangeListener(new PropertyChangeListener {
            def propertyChange(event: PropertyChangeEvent) {

              event.getPropertyName match {

                case "border" =>
                  if(event.getNewValue != null)
                    border = null

                case "opaque" =>
                  if(event.getNewValue.asInstanceOf[Boolean])
                    opaque = false

                case _ =>

              }
            }
          } )
        }

      // Close Button.
      if(closeable)
        contents += new Button {

          border = null
          icon = new ImageIcon(getClass.getResource("/icons/tab_close.png"))
          opaque = false

          listenTo(this)
          reactions += { case ButtonClicked(_) => pages.remove(t.index) }

          peer.addPropertyChangeListener(new PropertyChangeListener {
            def propertyChange(event: PropertyChangeEvent) {

              event.getPropertyName match {

                case "border" =>
                  if(event.getNewValue != null)
                  border = null

                case "opaque" =>
                  if(event.getNewValue.asInstanceOf[Boolean])
                    opaque = false

                case _ =>

              }
            }
          } )
        }
    } ).peer)
  }

  protected def rebuildTabBars { pages.foreach(t => overrideTabBar(t)) }

}

